namespace dnnl {
namespace impl {
namespace gpu {
namespace ocl {
const char *gen9_conv_nhwc_bwd_weights_f32_kernel[] ={ "/*******************************************************************************\n",
"* Copyright 2020 Intel Corporation\n",
"*\n",
"* Licensed under the Apache License, Version 2.0 (the \"License\");\n",
"* you may not use this file except in compliance with the License.\n",
"* You may obtain a copy of the License at\n",
"*\n",
"*     http://www.apache.org/licenses/LICENSE-2.0\n",
"*\n",
"* Unless required by applicable law or agreed to in writing, software\n",
"* distributed under the License is distributed on an \"AS IS\" BASIS,\n",
"* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n",
"* See the License for the specific language governing permissions and\n",
"* limitations under the License.\n",
"*******************************************************************************/\n",
"#pragma OPENCL EXTENSION cl_khr_int64_base_atomics : enable\n",
"#if ID > 1\n",
"#define CASE_3D 1\n",
"#else\n",
"#define CASE_3D 0\n",
"#endif\n",
"#define DIV_UP(a, b) (((a) + (b)-1) / (b))\n",
"#define RND_UP(a, b) (DIV_UP(a, b) * (b))\n",
"#if BWD_WEIGHTS == 1\n",
"inline void atomic_add_global(\n",
"        volatile __global atomic_float *source, float operand) {\n",
"    float old_val = atomic_load_explicit(\n",
"            source, memory_order_relaxed, memory_scope_device);\n",
"    if (isnan(operand)) return;\n",
"    bool success = false;\n",
"    do {\n",
"        float new_val = old_val + operand;\n",
"        success = atomic_compare_exchange_strong_explicit(source, &old_val,\n",
"                new_val, memory_order_acq_rel, memory_order_relaxed,\n",
"                memory_scope_device);\n",
"    } while (!success);\n",
"}\n",
"inline float read_ic_block(const __global float *ptr, int off) {\n",
"#if (IS_DW ? G : IC) % IC_BLOCK != 0\n",
"    int tail = (IS_DW ? G : IC) - off;\n",
"    if (tail < IC_BLOCK) {\n",
"        const int local_x = get_local_id(0);\n",
"        return (local_x < tail) ? ptr[local_x] : 0.0f;\n",
"    }\n",
"#endif\n",
"    return as_float(intel_sub_group_block_read((const __global uint *)ptr));\n",
"}\n",
"inline float read_oc_block(const __global float *ptr, int off) {\n",
"#if (IS_DW ? G : OC_WO_PADDING) % OC_BLOCK != 0\n",
"    int tail = (IS_DW ? G : OC_WO_PADDING) - off;\n",
"    if (tail < OC_BLOCK) {\n",
"        const int local_x = get_local_id(0);\n",
"        return (local_x < tail) ? ptr[local_x] : 0.0f;\n",
"    }\n",
"#endif\n",
"    return as_float(intel_sub_group_block_read((const __global uint *)ptr));\n",
"}\n",
"__attribute__((reqd_work_group_size(LWS_0, LWS_1, LWS_2))) // attr:no-format\n",
"__attribute__((intel_reqd_sub_group_size(SUB_GROUP_SIZE))) // attr:no-format\n",
"__kernel void\n",
"gen9_conv_nhwc_bwd_weights(__global float *src,\n",
"        volatile __global atomic_float *diff_wei,\n",
"        volatile __global atomic_float *diff_bias, __global float *diff_dst) {\n",
"    const int ksp = get_global_id(1);\n",
"#if CASE_3D\n",
"    const int kd = ksp / (KW * KH);\n",
"    const int khw = ksp % (KW * KH);\n",
"#else\n",
"    const int khw = ksp;\n",
"    const int kd = 0;\n",
"#endif\n",
"    const int kh = khw / KW;\n",
"    const int kw = khw % KW;\n",
"    const int local_x = get_local_id(0);\n",
"    const int chunk = get_global_id(2) % NCHUNK;\n",
"    const int icb_ocb = get_global_id(2) / NCHUNK;\n",
"    const int icb = icb_ocb % DIV_UP(IC, ICB);\n",
"    const int ocb = icb_ocb / DIV_UP(IC, ICB);\n",
"    const int ic_padded = RND_UP(IC, IC_BLOCK);\n",
"    const int oc_padded = RND_UP(OC, OC_BLOCK);\n",
"#if IS_DW\n",
"    const int g = 0;\n",
"    const int oc = get_group_id(0);\n",
"    const int ic = oc;\n",
"#else\n",
"    const int g_ic_oc = get_global_id(0);\n",
"    const int g = g_ic_oc / (oc_padded * DIV_UP(IC, IC_BLOCK));\n",
"    const int io = g_ic_oc % (oc_padded * DIV_UP(IC, IC_BLOCK));\n",
"    const int oc = (io % OCB) / OC_BLOCK + ocb * (OCB / OC_BLOCK);\n",
"    const int ic = (IC == 3) ? 0 : (io / OCB + icb * (ICB / IC_BLOCK));\n",
"#endif\n",
"    const int sp_chunk = chunk % OSP_CHUNK;\n",
"    const int mb_chunk = chunk / OSP_CHUNK;\n",
"    const int ow_nb = (OW + OWB - 1) / OWB;\n",
"    const int oh_nb = (OH + OHB - 1) / OHB;\n",
"    const int od_beg = ((sp_chunk / ow_nb) / oh_nb) * ODB;\n",
"    const int oh_beg = ((sp_chunk / ow_nb) % oh_nb) * OHB;\n",
"    const int ow_beg = (sp_chunk % ow_nb) * OWB;\n",
"    const int mb = mb_chunk * MB_CHUNK_SIZE;\n",
"    const int mb_end = min((mb_chunk + 1) * MB_CHUNK_SIZE, MB);\n",
"    const bool do_bias = (ic == 0 || IS_DW) && kh == 0 && kw == 0 && kd == 0;\n",
"    src += mb * ID * IH * IW * G * IC;\n",
"    src += g * IC + ic * IC_BLOCK;\n",
"    diff_dst += g * OC_WO_PADDING + oc * OC_BLOCK;\n",
"#if WITH_BIAS == 1\n",
"    diff_bias += g * OC_WO_PADDING + oc * OC_BLOCK + local_x;\n",
"    float bias_loc = 0.0f;\n",
"#endif\n",
"#if IC == 3\n",
"    float8 blockC00 = 0.0f;\n",
"#elif IS_DW\n",
"    float blockC00 = 0.0f;\n",
"#else\n",
"    float8 blockC00 = 0.0f;\n",
"    float8 blockC01 = 0.0f;\n",
"#endif\n",
"    for (int omb = mb; omb < mb_end; omb++) {\n",
"        const __global float *diff_dst1_\n",
"                = diff_dst + omb * OD * OH * OW * G * OC_WO_PADDING;\n",
"        for (int od = od_beg; od < min(od_beg + ODB, OD); od++)\n",
"            for (int oh = oh_beg; oh < min(oh_beg + OHB, OH); oh++) {\n",
"                const __global float *diff_dst1 = diff_dst1_\n",
"                        + (od * OH * OW + oh * OW) * G * OC_WO_PADDING;\n",
"                if (oh * SH + kh * (1 + DH) < PH\n",
"                        || oh * SH + kh * (1 + DH) >= IH + PH\n",
"#if CASE_3D\n",
"                        || od * SD + kd * (1 + DD) < PD\n",
"                        || od * SD + kd * (1 + DD) >= ID + PD\n",
"#endif\n",
"                ) {\n",
"#if WITH_BIAS == 1\n",
"                    if (do_bias) {\n",
"                        for (int ow = ow_beg; ow < ow_beg + OWB;\n",
"                                ow += OW_BLOCK) {\n",
"                            float8 blockB;\n",
"                            for (int i = 0; i < OW_BLOCK; i++) {\n",
"                                if (ow + i >= OW) {\n",
"                                    blockB[i] = 0.0;\n",
"                                } else {\n",
"                                    blockB[i] = read_oc_block(\n",
"                                            &diff_dst1[(ow + i) * G\n",
"                                                    * OC_WO_PADDING],\n",
"                                            oc * OC_BLOCK);\n",
"                                }\n",
"                            }\n",
"                            for (int i = 0; i < OW_BLOCK; i++)\n",
"                                bias_loc += blockB[i];\n",
"                        }\n",
"                    }\n",
"#endif\n",
"                    continue;\n",
"                }\n",
"                for (int ow = ow_beg; ow < ow_beg + OWB; ow += OW_BLOCK) {\n",
"                    const int id = od * SD - PD + kd * (1 + DD);\n",
"                    const int ih = oh * SH - PH + kh * (1 + DH);\n",
"                    const int iw = ow * SW - PW + kw * (1 + DW);\n",
"                    __global float *src1\n",
"                            = src + (id * IH * IW + ih * IW + iw) * G * IC;\n",
"#define TRANSPOSE_8(_block, _row, _col) \\\n",
"    { \\\n",
"        (float8)(intel_sub_group_shuffle(_block[_row], 0 + _col), \\\n",
"                intel_sub_group_shuffle(_block[_row], 1 + _col), \\\n",
"                intel_sub_group_shuffle(_block[_row], 2 + _col), \\\n",
"                intel_sub_group_shuffle(_block[_row], 3 + _col), \\\n",
"                intel_sub_group_shuffle(_block[_row], 4 + _col), \\\n",
"                intel_sub_group_shuffle(_block[_row], 5 + _col), \\\n",
"                intel_sub_group_shuffle(_block[_row], 6 + _col), \\\n",
"                intel_sub_group_shuffle(_block[_row], 7 + _col)) \\\n",
"    }\n",
"#define FMA8(a, b, c) fma((float8)(a), (float8)b, (float8)c)\n",
"#define MULTIPLY_BLOCKS_8x8(_result, _blockA, _blockB, col) \\\n",
"    { \\\n",
"        _result = FMA8(_blockB.s0, TRANSPOSE_8(_blockA, 0, col), _result); \\\n",
"        _result = FMA8(_blockB.s1, TRANSPOSE_8(_blockA, 1, col), _result); \\\n",
"        _result = FMA8(_blockB.s2, TRANSPOSE_8(_blockA, 2, col), _result); \\\n",
"        _result = FMA8(_blockB.s3, TRANSPOSE_8(_blockA, 3, col), _result); \\\n",
"        _result = FMA8(_blockB.s4, TRANSPOSE_8(_blockA, 4, col), _result); \\\n",
"        _result = FMA8(_blockB.s5, TRANSPOSE_8(_blockA, 5, col), _result); \\\n",
"        _result = FMA8(_blockB.s6, TRANSPOSE_8(_blockA, 6, col), _result); \\\n",
"        _result = FMA8(_blockB.s7, TRANSPOSE_8(_blockA, 7, col), _result); \\\n",
"    }\n",
"                    float8 blockA, blockB;\n",
"#if IC == 3\n",
"                    if (local_x < IC) {\n",
"                        for (int i = 0; i < OW_BLOCK; i++) {\n",
"                            if (iw + i * SW < 0 || iw + i * SW >= IW) {\n",
"                                blockA[i] = 0;\n",
"                            } else {\n",
"                                blockA[i] = src1[i * SW * G * IC + local_x];\n",
"                            }\n",
"                        }\n",
"                    } else {\n",
"                        blockA = 0.0f;\n",
"                    }\n",
"#else\n",
"                    __attribute__((opencl_unroll_hint(8))) // attr:no-format\n",
"                    for (int i = 0; i < OW_BLOCK; i++) {\n",
"                        if (iw + i * SW < 0 || iw + i * SW >= IW) {\n",
"                            blockA[i] = 0;\n",
"                        } else {\n",
"                            blockA[i] = read_ic_block(\n",
"                                    &src1[i * SW * G * IC], ic * IC_BLOCK);\n",
"                        }\n",
"                    }\n",
"#endif\n",
"                    __attribute__((opencl_unroll_hint(8))) // attr:no-format\n",
"                    for (int i = 0; i < OW_BLOCK; i++) {\n",
"                        if (ow + i >= OW) {\n",
"                            blockB[i] = 0.0;\n",
"                        } else {\n",
"                            blockB[i] = read_oc_block(\n",
"                                    &diff_dst1[(ow + i) * G * OC_WO_PADDING],\n",
"                                    oc * OC_BLOCK);\n",
"                        }\n",
"                    }\n",
"#if IC == 3\n",
"                    MULTIPLY_BLOCKS_8x8(blockC00, blockA, blockB, 0);\n",
"#elif IS_DW\n",
"                    for (int i = 0; i < OW_BLOCK; i++) {\n",
"                        blockC00 = fma(blockA[i], blockB[i], blockC00);\n",
"                    }\n",
"#else\n",
"                    MULTIPLY_BLOCKS_8x8(blockC00, blockA, blockB, 0);\n",
"                    MULTIPLY_BLOCKS_8x8(blockC01, blockA, blockB, 8);\n",
"#endif\n",
"#if WITH_BIAS == 1\n",
"                    for (int i = 0; i < 8; i++)\n",
"                        bias_loc += blockB[i];\n",
"#endif\n",
"                }\n",
"            }\n",
"        src += ID * IH * IW * G * IC;\n",
"    }\n",
"#if WITH_BIAS == 1\n",
"    if (do_bias && oc * OC_BLOCK + local_x < (IS_DW ? G : OC_WO_PADDING))\n",
"        atomic_add_global(diff_bias, bias_loc);\n",
"#endif\n",
"#if IC == 3\n",
"    diff_wei += g * oc_padded * ic_padded * KD * KH * KW;\n",
"    diff_wei += oc * KD * KH * KW * ic_padded * OC_BLOCK;\n",
"    diff_wei += (kd * KH * KW + kh * KW + kw) * ic_padded * OC_BLOCK;\n",
"    for (int i = 0; i < 3; i++)\n",
"        atomic_add_global(diff_wei + i * OC_BLOCK + local_x, blockC00[i]);\n",
"#elif IS_DW\n",
"    diff_wei += oc * KD * KH * KW * OC_BLOCK;\n",
"    diff_wei += (kd * KH * KW + kh * KW + kw) * OC_BLOCK;\n",
"    atomic_add_global(diff_wei + local_x, blockC00);\n",
"#else\n",
"    diff_wei += g * ic_padded * oc_padded * KD * KH * KW;\n",
"    diff_wei += ic * oc_padded * KD * KH * KW * IC_BLOCK;\n",
"    diff_wei += oc * KD * KH * KW * IC_BLOCK * OC_BLOCK;\n",
"    diff_wei += (kd * KH * KW + kh * KW + kw) * IC_BLOCK * OC_BLOCK;\n",
"    for (int i = 0; i < 8; i++)\n",
"        atomic_add_global(diff_wei + i * OC_BLOCK + local_x, blockC00[i]);\n",
"    for (int i = 0; i < 8; i++)\n",
"        atomic_add_global(diff_wei + (8 + i) * OC_BLOCK + local_x, blockC01[i]);\n",
"#endif\n",
"}\n",
"#endif", nullptr };
}
}
}
}