namespace dnnl {
namespace impl {
namespace gpu {
namespace ocl {
const char *ref_zero_pad_kernel[] ={ "/*******************************************************************************\n",
" * Copyright 2020-2021 Intel Corporation\n",
" *\n",
" * Licensed under the Apache License, Version 2.0 (the \"License\");\n",
" * you may not use this file except in compliance with the License.\n",
" * You may obtain a copy of the License at\n",
" *\n",
" *     http://www.apache.org/licenses/LICENSE-2.0\n",
" *\n",
" * Unless required by applicable law or agreed to in writing, software\n",
" * distributed under the License is distributed on an \"AS IS\" BASIS,\n",
" * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n",
" * See the License for the specific language governing permissions and\n",
" * limitations under the License.\n",
" *******************************************************************************/\n",
"// This define is required before including zero_pad_struct.h to allow reuse of\n",
"// the same structure with the correct data types between OpenCL kernels and C++\n",
"// code.\n",
"#define IS_OCL_KERNEL\n",
"/*******************************************************************************\n",
" * Copyright 2020-2021 Intel Corporation\n",
" *\n",
" * Licensed under the Apache License, Version 2.0 (the \"License\");\n",
" * you may not use this file except in compliance with the License.\n",
" * You may obtain a copy of the License at\n",
" *\n",
" *     http://www.apache.org/licenses/LICENSE-2.0\n",
" *\n",
" * Unless required by applicable law or agreed to in writing, software\n",
" * distributed under the License is distributed on an \"AS IS\" BASIS,\n",
" * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n",
" * See the License for the specific language governing permissions and\n",
" * limitations under the License.\n",
" *******************************************************************************/\n",
"#ifndef GPU_ZERO_PAD_ZERO_PAD_STRUCT_H\n",
"#define GPU_ZERO_PAD_ZERO_PAD_STRUCT_H\n",
"#define ZERO_PAD_MAX_STEP_SIZE 1536\n",
"#ifdef IS_OCL_KERNEL\n",
"#define ZERO_PAD_MASK_DATA_TYPE uchar\n",
"#else\n",
"#define ZERO_PAD_MASK_DATA_TYPE unsigned char\n",
"#endif\n",
"#define ZERO_PAD_MASK_DT_BITS (8 * sizeof(ZERO_PAD_MASK_DATA_TYPE))\n",
"#define ZERO_PAD_MASK_SIZE (ZERO_PAD_MAX_STEP_SIZE / ZERO_PAD_MASK_DT_BITS)\n",
"#define ZERO_PAD_BIT_MODE 0\n",
"#define ZERO_PAD_LOOKUP_MODE 1\n",
"typedef struct {\n",
"    ZERO_PAD_MASK_DATA_TYPE mask[ZERO_PAD_MASK_SIZE];\n",
"} zero_pad_mask_t;\n",
"#endif\n",
"#define DEFAULT_NELEMS_BLOCK 8\n",
"static inline void typed_ref_zero_pad(__global void *a, ulong type_size,\n",
"        ulong step_nelems, ulong nelems_block, ulong step_block, ulong nsteps,\n",
"        ulong step_size, zero_pad_mask_t step_bitmask, ulong mode) {\n",
"    const int i0 = get_global_id(0);\n",
"    const int istep = get_global_id(1) * step_block;\n",
"    const int iblock = get_global_id(2);\n",
"    int offset = iblock * step_size + (step_size - nsteps * step_nelems)\n",
"            + istep * step_nelems;\n",
"    const int step = ZERO_PAD_MASK_DT_BITS;\n",
"    // Interpret buffer differently based on type_size, this is implicitly\n",
"    // using the fact that the bit representation of 0 is the same regardless of\n",
"    // data type, e.g. int32 and f32 represent 0 the same.\n",
"    __global int *a4 = (__global int *)a;\n",
"    __global short *a2 = (__global short *)a;\n",
"    __global char *a1 = (__global char *)a;\n",
"    if (mode == ZERO_PAD_BIT_MODE) {\n",
"        // Use a bit mask to determine which elements to zero in a block\n",
"        // This allows for handling larger blocks than the lookup mode without\n",
"        // increasing the size of zero_pad_mask_t. There is a disadvantage as\n",
"        // more time is spent checking if a given piece of memory needs zeroed\n",
"        for (int k = 0; k < step_block; k++) {\n",
"            __attribute__((opencl_unroll_hint)) // attr:no-format\n",
"            for (int i = i0; i < step_nelems; i += nelems_block) {\n",
"                if (step_bitmask.mask[i / step] & (1 << (i % step))) {\n",
"                    switch (type_size) {\n",
"                        case 4: a4[offset + i] = 0; break;\n",
"                        case 2: a2[offset + i] = 0; break;\n",
"                        case 1: a1[offset + i] = 0; break;\n",
"                    }\n",
"                }\n",
"            }\n",
"            offset += step_nelems;\n",
"        }\n",
"    } else {\n",
"        // mode == ZERO_PAD_LOOKUP_MODE\n",
"        // Use a lookup table to determine which elements to zero in a block\n",
"        // Implementation requires global_work_size(0) is equal to the number of\n",
"        // elements to zero in a given block\n",
"        int i = step_bitmask.mask[i0];\n",
"        for (int k = 0; k < step_block; k++) {\n",
"            switch (type_size) {\n",
"                case 4: a4[offset + i] = 0; break;\n",
"                case 2: a2[offset + i] = 0; break;\n",
"                case 1: a1[offset + i] = 0; break;\n",
"            }\n",
"            offset += step_nelems;\n",
"        }\n",
"    }\n",
"}\n",
"static inline void sized_ref_zero_pad(__global void *a, ulong type_size,\n",
"        ulong step_nelems, ulong nelems_block, ulong step_block, ulong nsteps,\n",
"        ulong step_size, zero_pad_mask_t step_bitmask, ulong mode) {\n",
"    switch (type_size) {\n",
"        case 4:\n",
"            typed_ref_zero_pad((__global float *)a, 4, step_nelems,\n",
"                    nelems_block, step_block, nsteps, step_size, step_bitmask,\n",
"                    mode);\n",
"            break;\n",
"        case 2:\n",
"            typed_ref_zero_pad((__global float *)a, 2, step_nelems,\n",
"                    nelems_block, step_block, nsteps, step_size, step_bitmask,\n",
"                    mode);\n",
"            break;\n",
"        case 1:\n",
"            typed_ref_zero_pad((__global float *)a, 1, step_nelems,\n",
"                    nelems_block, step_block, nsteps, step_size, step_bitmask,\n",
"                    mode);\n",
"            break;\n",
"    }\n",
"}\n",
"__kernel void ref_zero_pad(__global void *a, ulong type_size, ulong step_nelems,\n",
"        ulong nelems_block, ulong step_block, ulong nsteps, ulong step_size,\n",
"        zero_pad_mask_t step_bitmask, ulong mode) {\n",
"    // Use a switch statement here to allow constant propagation optimizations to\n",
"    // remove switch statement from loop in typed_ref_zero_pad.\n",
"    switch (step_nelems) {\n",
"        case 16:\n",
"            sized_ref_zero_pad(a, type_size, 16, DEFAULT_NELEMS_BLOCK,\n",
"                    step_block, nsteps, step_size, step_bitmask, mode);\n",
"            break;\n",
"        case 32:\n",
"            sized_ref_zero_pad(a, type_size, 32, DEFAULT_NELEMS_BLOCK,\n",
"                    step_block, nsteps, step_size, step_bitmask, mode);\n",
"            break;\n",
"        case 64:\n",
"            sized_ref_zero_pad(a, type_size, 64, DEFAULT_NELEMS_BLOCK,\n",
"                    step_block, nsteps, step_size, step_bitmask, mode);\n",
"            break;\n",
"        default:\n",
"            sized_ref_zero_pad(a, type_size, step_nelems, nelems_block,\n",
"                    step_block, nsteps, step_size, step_bitmask, mode);\n",
"            break;\n",
"    }\n",
"}\n",
"__attribute__((intel_reqd_sub_group_size(16))) __kernel void\n",
"ref_zero_pad_subg_16(__global void *a, const uint type_size,\n",
"        const ulong base_offset, const ulong b_block_size,\n",
"        const ulong b_block_offset, const ulong d0_stride,\n",
"        const ulong d1_stride, const ulong d2_stride, const ulong d3_stride,\n",
"        const unsigned d0_size, const unsigned d1_size, const unsigned d2_size,\n",
"        const unsigned d3_size, const uint b_multiplier) {\n",
"    const unsigned a_block_id = get_global_id(0) / 16;\n",
"    const unsigned b_block_id = get_global_id(1);\n",
"    unsigned mixed_dims = get_global_id(2);\n",
"    const unsigned d3_dim = mixed_dims % d3_size;\n",
"    mixed_dims /= d3_size;\n",
"    const unsigned d2_dim = mixed_dims % d2_size;\n",
"    mixed_dims /= d2_size;\n",
"    const unsigned d1_dim = mixed_dims % d1_size;\n",
"    const unsigned d0_dim = mixed_dims / d1_size;\n",
"    __global void *p = a + base_offset;\n",
"    p += a_block_id * b_block_size;\n",
"    p += b_block_id * b_block_offset;\n",
"    p += d0_dim * d0_stride;\n",
"    p += d1_dim * d1_stride;\n",
"    p += d2_dim * d2_stride;\n",
"    p += d3_dim * d3_stride;\n",
"    const unsigned stride = 16 * type_size;\n",
"    for (unsigned b_midx = 0; b_midx < b_multiplier; ++b_midx) {\n",
"        switch (type_size) {\n",
"            case 4: intel_sub_group_block_write((__global uint *)p, 0); break;\n",
"            case 2:\n",
"                intel_sub_group_block_write_us((__global ushort *)p, 0);\n",
"                break;\n",
"            case 1:\n",
"                intel_sub_group_block_write_uc((__global uchar *)p, 0);\n",
"                break;\n",
"        }\n",
"        p += stride;\n",
"    }\n",
"}\n",
"__attribute__((intel_reqd_sub_group_size(16))) __kernel void\n",
"ref_zero_pad_subg_16_mask_and_clear_dt_1b(__global void *a, const uint mask) {\n",
"    const uint block_size = 8;\n",
"    const uint data_stride = 32;\n",
"    const uint simd = 16;\n",
"    const ulong offset = get_global_id(0) * block_size;\n",
"    const unsigned subg_local_id = get_sub_group_local_id();\n",
"    __global void *p = a + offset;\n",
"    const uint mask_val = mask > subg_local_id ? 1 : 0;\n",
"    uchar val_c[block_size];\n",
"    for (unsigned idx = 0; idx < block_size / 2; ++idx) {\n",
"        val_c[idx * 2] = intel_sub_group_block_read_uc(\n",
"                (__global uchar *)(p + data_stride * idx));\n",
"    }\n",
"    for (unsigned idx = 1; idx < block_size; idx += 2) {\n",
"        val_c[idx] = 0;\n",
"    }\n",
"    for (unsigned idx = 0; idx < block_size; idx += 2) {\n",
"        val_c[idx] *= (uchar)mask_val;\n",
"    }\n",
"    for (unsigned idx = 0; idx < block_size; idx += 8) {\n",
"        intel_sub_group_block_write_uc8(\n",
"                (__global uchar *)(p + simd * idx), *((uchar8 *)(&val_c[idx])));\n",
"    }\n",
"}", nullptr };
}
}
}
}